package Class::Autouse;

# This is a copy of Class::Autouse 1.03 with POD removed

### Memory Overhead: 196K

require 5.005;
use strict;
no strict 'refs';
use UNIVERSAL ();

# Become an exporter so we don't get
# complaints when we act as a pragma.
use base 'Exporter';

# There are some things we use
use Carp       ();
use File::Spec ();
use List::Util ();

# Globals
use vars qw{$VERSION $DEBUG};
use vars qw{$devel $superloader};
use vars qw{%chased %loaded %special %bad};
use vars qw{*_original_can};
BEGIN {
	$VERSION = '1.03';
	$DEBUG   = 0;

	# Using Class::Autouse in a mod_perl situation can be dangerous, 
	# as it can clash badly with autoreloaders such as Apache::Reload.
	# So ALWAYS run in devel mode under mod_perl. Since they should
	# probably be loading modules at startup time ( in the parent 
	# process ) anyway, this is a good thing.
	$devel       = !! $ENV{MOD_PERL};
	$superloader = 0;

	# Anti-loop protection. "Have we tried to autoload a method before?"
	# Contains the fully referenced sub name.
	%chased = ();

	# ERRATA
	# Special classes are internal and should be left alone.
	# Loaded modules are those we should assume to be loaded already.
	# Bad classes are those that are irrepairably incompatible with us.
	%special = map { $_ => 1 } qw{CORE main ARRAY HASH SCALAR REF UNIVERSAL};
	%loaded  = map { $_ => 1 } qw{UNIVERSAL Exporter Carp File::Spec};
	%bad     = map { $_ => 1 } qw{IO::File};
}





#####################################################################
# Configuration and Setting up

# Developer mode flag.
# Don't let them turn it off if we are under mod_perl.
sub devel {
	_debug(\@_, 1) if $DEBUG;

	$devel = ($_[1] and ! $ENV{MOD_PERL}) ? 1 : 0;
}

# Happy Fun Super Loader!
# The process here is to replace the &UNIVERSAL::AUTOLOAD sub
# ( which is just a dummy by default ) with a flexible class loader.
sub superloader {
	_debug(\@_ ,1) if $DEBUG;

	return 1 if $superloader;

	# Overwrite UNIVERSAL::AUTOLOAD and also catch any
	# DESTROY calls that make it to UNIVERSAL, so they
	# don't trigger an AUTOLOAD call.
	*UNIVERSAL::AUTOLOAD = \&_autoload;
	*UNIVERSAL::DESTROY = \&_destroy;
	$superloader = 1;
}

# The main autouse sub
sub autouse {
	# Remove any reference to ourselves, to allow us to
	# operate as a function, or a method
	shift if $_[0] eq 'Class::Autouse';
	return 1 unless @_;

	_debug(\@_) if $DEBUG;

	my @classes = grep { $_ } @_;
	foreach my $class ( @classes ) {
		# Control flag handling
		if ( $class =~ s/^:// ) {
			if ( $class eq 'superloader' ) {
				# Turn on the superloader
				Class::Autouse->superloader();
			} elsif ( $class eq 'devel' ) {
				# Turn on devel mode
				Class::Autouse->devel( 1 );
			} elsif ( $class eq 'debug' ) {
				# Turn on debugging
				$DEBUG = 1;
				print _call_depth(1) . "Class::Autouse::autoload -> Debugging Activated.\n";
			}
			next;
		}

		# Load now if in devel mode, or if it's a bad class
		if ( $devel || $bad{$class} ) {
			Class::Autouse->load( $class );
			next;
		}

		# Get the file name
		my $file = File::Spec->catfile( split( /::/, $class) ) . '.pm';

		# Does the file for the class exist?
		next if exists $INC{$file};
		unless ( _file_exists($file) ) {
			_cry( "Can't locate $file in \@INC (\@INC contains: @INC)" );
		}

		# Don't actually do anything if the superloader is on
		next if $superloader;

		# Add the AUTOLOAD hook and %INC lock to prevent 'use'ing
		*{"${class}::AUTOLOAD"} = \&_autoload;
		$INC{$file} = 'Class::Autouse';
	}

	1;
}

# Link import to autouse, so we can act as a pragma
BEGIN {
	*import = *autouse;
}





#####################################################################
# Explicit Actions

# Completely load a class ( The class and all it's dependencies ).
sub load {
	_debug(\@_, 1) if $DEBUG;

	my $class = $_[1] or _cry( "No class name specified to load" );
	return 1 if $loaded{$class};

	# Load the entire ISA tree
	my @stack  = ( $class );
	my %seen   = ( 'UNIVERSAL' => 1 );
	my @search = ();
	while ( my $c = shift @stack ) {
		next if $seen{$c}++;

		# Ensure class is loaded
		_load($c) unless $loaded{$c};

		# Add the class to the search list,
		# and add the @ISA to the load stack.
		push @search, $c;
        	unshift @stack, @{"${c}::ISA"};
	}

	# If called an an array context, return the ISA tree.
	# In scalar context, just return true.
	wantarray ? @search : 1;
}

# Is a particular class installed in out @INC somewhere
# OR is it loaded in our program already
sub class_exists {
	_debug(\@_, 1) if $DEBUG;

	# Is the class loaded already, or can we find it's file
	_namespace_occupied($_[1]) or _file_exists($_[1]);
}

# A more general method to answer the question
# "Can I call a method on this class and expect it to work"
# Returns undef if the class does not exist
# Returns 0 if the class is not loaded ( or autouse'd )
# Returns 1 if the class can be used.
sub can_call_methods {
	_debug(\@_, 1) if $DEBUG;

	# Is it loaded already, or is the file in %INC
	_namespace_occupied( $_[1] ) or exists $INC{ File::Spec->catfile( split( /::/, $_[1]) ) . '.pm' };
}

# Recursive methods currently only work withing the scope of the single @INC
# entry containing the "top" module, and will probably stay this way

# Autouse not only a class, but all others below it.
sub autouse_recursive {
	_debug(\@_, 1) if $DEBUG;

	# Just load if in devel mode
	my $class = $_[1];
	return Class::Autouse->load_recursive( $class ) if $devel;

	# Don't need to do anything if the super loader is on
	return 1 if $superloader;

	# Find all the child classes, and hand them to the autouse method
	Class::Autouse->autouse( $class, _child_classes($class) );
}

# Load not only a class and all others below it
sub load_recursive {
	_debug(\@_, 1) if $DEBUG;

	# Load the parent class, and it's children
	my $class = $_[1];
	foreach ( $class, _child_classes($class) ) {
		Class::Autouse->load($_);
	}

	1;
}





#####################################################################
# Symbol Table Hooks

# These get hooked to various places on the symbol table,
# to enable the autoload functionality

# Get's linked via the symbol table to any AUTOLOADs are required
sub _autoload {
	_debug(\@_, 0, ", AUTOLOAD = '$Class::Autouse::AUTOLOAD'") if $DEBUG;

	# Loop detection ( Just in case )
	my $method = $Class::Autouse::AUTOLOAD or _cry( "Missing method name" );
	_cry( "Undefined subroutine &$method called" ) if ++$chased{ $method } > 10;

	# Don't bother with special classes
	my ($class, $function) = $method =~ m/^(.*)::(.*)$/so;
	_cry( "Undefined subroutine \&$method called" ) if $special{$class};

	# Load the class and it's dependancies, and get the search path
	my @search = Class::Autouse->load($class);

	# Find and go to the named method
	my $found = List::Util::first { defined *{"$_\::$function"}{CODE} } @search;
	goto &{"${found}::$function"} if $found;

	# Check for package AUTOLOADs
	foreach my $c ( @search ) {
        	if ( defined *{ "${c}::AUTOLOAD" }{CODE} ) {
			# Simulate a normal autoload call
        		${"${c}::AUTOLOAD"} = $method;
        		goto &{"${c}::AUTOLOAD"};
        	}
	}

	# Can't find the method anywhere. Throw the same error Perl does.
	_cry( "Can't locate object method \"$function\" via package \"$class\"" );
}

# This just handles the call and does nothing
sub _destroy { _debug(\@_) if $DEBUG }

# This is the replacement for UNIVERSAL::can
sub _can {
	my $class = ref $_[0] || $_[0] || return undef;

	# Shortcut for the most likely cases
	goto &_original_can if $loaded{$class} or defined @{"${class}::ISA"};

	# Does it look like a package?
	return undef unless $class =~ qr/^[^\W\d]\w*(?:(?:'|::)[^\W\d]\w*)*$/;

	# Do we try to load the class
	my $load = 0;
	my $file = File::Spec->catfile( split( /::/, $class) ) . '.pm';
	if ( defined $INC{$file} and $INC{$file} eq 'Class::Autouse' ) {
		# It's an autoused class
		$load = 1;
	} elsif ( ! $superloader ) {
		# Superloader isn't on, don't load
		$load = 0;
	} elsif ( _namespace_occupied($class) ) {
		# Superloader is on, but there is something already in the class
		# This can't be the autouse loader, because we would have caught
		# that case already
		$load = 0;
	} else {
		# The rules of the superloader say we assume loaded unless we can
		# tell otherwise. Thus, we have to have a go at loading.
		$load = 1;
	}

	# If needed, load the class and all it's dependencies.
	# UNIVERSAL::can never dies, so we shouldn't either.
	# Ignore, errors. If something goes wrong,
	# let the real UNIVERSAL::can have a short at it anyway.
	eval { Class::Autouse->load($class) } if $load;

	# Hand off to the real UNIVERSAL::can
	goto &_original_can;
}





#####################################################################
# Support Functions

# Load a single class
sub _load {
	_debug(\@_, 1) if $DEBUG;

	# Don't attempt to load special classes
	my $class = shift or _cry( "Did not specify a class to load" );
	return 1 if $special{$class};

	# Run some checks
	my $file = File::Spec->catfile( split( /::/, $class) ) . '.pm';
	if ( defined $INC{$file} ) {
		# If the %INC lock is set to any other value, the file is 
		# already loaded. We do not need to do anything.
		return $loaded{$class} = 1 if $INC{$file} ne 'Class::Autouse';

		# Because we autoused it earlier, we know the file for this
		# class MUST exist.
		# Removing the AUTOLOAD hook and %INC lock is all we have to do
		delete ${"${class}::"}{'AUTOLOAD'};
		delete $INC{ $file };

	} elsif ( ! _file_exists($file) ) {
		# File doesn't exist. We might still be OK, if the class was
		# defined in some other module that got loaded a different way.
		_namespace_occupied( $class )
			? return $loaded{$class} = 1
			: _cry( "Can't locate $file in \@INC (\@INC contains: @INC)" );
	}

	# Load the file
	print _call_depth(1) . "  Class::Autouse::load -> Loading in $file\n" if $DEBUG;
	eval { require $file }; _cry($@) if $@;
	$loaded{$class} = 1;
}

# Find all the child classes for a parent class.
# Returns in the list context.
sub _child_classes {
	_debug(\@_) if $DEBUG;

	# Get the classes file name
	my $base_class = shift;
	my $base_file = File::Spec->catfile( split( /::/, $base_class) ) . '.pm';

	# Find where it is in @INC
	my $inc_path = List::Util::first {
		-f File::Spec->catfile( $_, $base_file )
		} @INC or return;

	# Does the file have a subdirectory
	# i.e. Are there child classes
	my $child_path = substr( $base_file, 0, length($base_file) - 3 );
	my $child_path_full = File::Spec->catdir( $inc_path, $child_path );
	return 0 unless -d $child_path_full and -r $child_path_full;

	# Main scan loop
	my ( $dir, @files );
	my @modules = ();
	my @queue = ( $child_path );
	while ( $dir = pop @queue ) {
		my $full_dir = File::Spec->catdir( $inc_path, $dir );

		# Read in the raw file list
		# Skip directories we can't open
		opendir( FILELIST, $full_dir ) or next;
		@files = readdir FILELIST;
		closedir FILELIST;

		# Iterate over them
		@files = map { File::Spec->catfile( $dir, $_ ) } # Full relative path
			grep { ! /^\./ } @files;                 # Ignore hidden files
		foreach my $file ( @files ) {
			my $full_file = File::Spec->catfile( $inc_path, $file );

			# Add to the queue if it's a directory we can descend
			if ( -d $full_file and -r $full_file ) {
				push @queue, $file;
				next;
			}

			# We only want .pm files we can read
			next unless substr( $file, length($file) - 3 ) eq '.pm';
			next unless -f $full_file;

			push @modules, $file;
		}
	}

	# Convert the file names into modules
	map { join '::', File::Spec->splitdir($_) }
		map { substr( $_, 0, length($_) - 3 ) } @modules;
}





#####################################################################
# Private support methods

# Does a class or file exists somewhere in our include path. For 
# convenience, returns the unresolved file name ( even if passed a class )
sub _file_exists {
	_debug(\@_) if $DEBUG;

	# What are we looking for?
	my $file = shift or return undef;
	return undef if $file =~ m/(?:\012|\015)/o;

	# If provided a class name, convert it
	if ( $file =~ /::/o ) {
		$file = File::Spec->catfile( split( /::/, $file) ) . '.pm';
	}

	# Scan @INC for the file
	foreach ( @INC ) {
		return $file if -f File::Spec->catfile( $_, $file );
	}

	undef;
}

# Is a namespace occupied by anything significant
sub _namespace_occupied {
	_debug(\@_) if $DEBUG;

	# Handle the most likely case
	my $class = shift or return undef;
	return 1 if defined @{"${class}::ISA"};

	# Get the list of glob names
	foreach ( keys %{"${class}::"} ) {
		# Only check for methods, since that's all that's reliable
		return 1 if defined *{"${class}::$_"}{CODE};
	}

	'';
}

# Establish our call depth
sub _call_depth {
	my $spaces = shift;
	if ( $DEBUG and ! $spaces ) { _debug(\@_); }

	# Search up the caller stack to find the first call that isn't us.
	my $level = 0;
	while( $level++ < 1000 ) {
		my @call = caller( $level );
		my ($subclass) = $call[3] =~ m/^(.*)::/so;
		unless ( defined $subclass and $subclass eq 'Class::Autouse' ) {
			# Subtract 1 for this sub's call
			$level -= 1;
			return $spaces ? join( '', (' ') x ($level - 2)) : $level;
		}
	}

	Carp::croak( "Infinite loop trying to find call depth" );
}

# Die gracefully
sub _cry {
	_debug() if $DEBUG;

	local $Carp::CarpLevel;
	$Carp::CarpLevel += _call_depth();
	Carp::croak( $_[0] );
}

# Adaptive debug print generation
sub _debug {
	my $args = shift;
	my $method = !! shift;
	my $message = shift || '';
	my @c = caller(1);
	my $msg = _call_depth(1) . $c[3];
	if ( ref $args ) {
		my @mapped = map { "'$_'" } @$args;
		shift @mapped if $method;
		$msg .= @mapped ? "( " . ( join ', ', @mapped ) . " )" : '()';
	}

	print $msg . $message . "\n";
}





#####################################################################
# Final Initialisation

# Replace the UNIVERSAL::can sub with our own version
BEGIN {
	# We don't need to do this if we are forced into devel mode
	# by the whole mod_perl situation
	unless ( $ENV{MOD_PERL} ) {
		# Copy the current UNIVERSAL::can
		*_original_can = *UNIVERSAL::can{CODE};
	
		# Replace UNIVERSAL::can, using different methods of disabling warnings
		# depending if we have access to the warnings module or not.
		if ( class_exists('warnings') ) {
			no warnings;
			*UNIVERSAL::can = *Class::Autouse::_can{CODE};
		} else {
			local $^W = 0;
			*UNIVERSAL::can = *Class::Autouse::_can{CODE};
		}
	}
}

1;
